#!/bin/bash

# This product is unlike anything else we build 

usage()
{
   echo "USAGE: `basename ${0}` <product_dir> <e6> [tar]"
}

# -------------------------------------------------------------------
# shared boilerplate
# -------------------------------------------------------------------

get_this_dir() 
{
    ( cd / ; /bin/pwd -P ) >/dev/null 2>&1
    if (( $? == 0 )); then
      pwd_P_arg="-P"
    fi
    reldir=`dirname ${0}`
    thisdir=`cd ${reldir} && /bin/pwd ${pwd_P_arg}`
}

get_ssibuildshims()
{
    # make sure we can use the setup alias
    if [ -z ${UPS_DIR} ]
    then
       echo "ERROR: please setup ups"
       exit 1
    fi
    source `${UPS_DIR}/bin/ups setup ${SETUP_UPS}`
    
    setup ssibuildshims ${ssibuildshims_version} -z ${product_dir}:$PRODUCTS
}

set_qualifiers()
{
    if [ -z ${basequal} ]
    then
       echo "ERROR: please specify the qualifiers"
       usage
       exit 1
    fi
    qualdir=`echo ${basequal} | sed -e 's/:/-/g'`
    fullqual=${basequal}
}

get_platform_name()
{
    OS1=`uname`
    flvr=`ups flavor`
    if [ ${OS1} = "Darwin" ]
    then
	flvr=`ups flavor -2`
    fi
    OS=`${SSIBUILDSHIMS_DIR}/bin/get_os_platform`
    plat=`${SSIBUILDSHIMS_DIR}/bin/get_os_platform osplat`

    if [ -z ${flvr} ]
    then
       echo "ERROR: could not determine flavor for this platform"
       echo "       something is very wrong"
       exit 1
    fi

    pkgdir=${product_dir}/${package}/${pkgver}/${flvr}-${qualdir}
    patchdir=${product_dir}/${package}/${pkgver}/patch
    blddir=${product_dir}/${package}/${pkgver}
}

make_distribution_tarball()
{
    cd ${product_dir}
    setup cetpkgsupport || die "cannot setup cetpkgsupport"
    set -x
    tar -cjf ${package}-${pkgdotver}-`get-directory-name os`-`get-directory-name platform`-${qualdir}.tar.bz2 \
        ${package}/${pkgver}.version/${flvr}_${basequal} \
        ${package}/${pkgver}/ups \
        ${package}/${pkgver}/doc \
        ${package}/${pkgver}/include \
        ${package}/${pkgver}/module \
        ${package}/${pkgver}/script \
        ${package}/${pkgver}/src_* \
        ${package}/${pkgver}/README \
        ${package}/${pkgver}/Makefile \
        ${package}/${pkgver}/build_trace.sh \
        ${package}/${pkgver}/${flvr}-${qualdir} 
    set +x
}

die()
{
  echo "$1"; exit 2
}

# -------------------------------------------------------------------
# start processing
# -------------------------------------------------------------------

product_dir=${1}
basequal=${2}
maketar=${3}

if [ -z ${product_dir} ]
then
   echo "ERROR: please specify the local product directory"
   usage
   exit 1
fi

# -------------------------------------------------------------------
# package name and version
# -------------------------------------------------------------------

package=TRACE
origpkgver=v3_04_00
pkgver=${origpkgver}
pkgdotver=`echo ${pkgver} | sed -e 's/_/./g' | sed -e 's/^v//'`
ssibuildshims_version=v0_08_02

get_this_dir

get_ssibuildshims

set_qualifiers
 
get_platform_name

if [ "${maketar}" = "tar" ] && [ -d ${pkgdir}/lib ]
then
   make_distribution_tarball
   exit 0
fi

echo "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${flvr})"

# -------------------------------------------------------------------
# package dependent stuff goes here
# -------------------------------------------------------------------

mkdir -p ${pkgdir}
if [ ! -d ${pkgdir} ]
then
   echo "ERROR: failed to create ${pkgdir}"
   exit 1
fi
mkdir -p ${blddir}
if [ ! -d ${blddir} ]
then
   echo "ERROR: failed to create ${blddir}"
   exit 1
fi

# apply patch for ups file

cd ${blddir}
patch -b -p0 < "${patchdir}/trace-${origpkgver}.patch" 
status=$?
echo "result of patch is ${status}"
#|| die "ERROR: application of patch ${patchdir}/trace-${origpkgver}.patch failed."

# declare now so we can setup
# fake ups declare
fakedb=${product_dir}/${package}/${pkgver}/fakedb
${SSIBUILDSHIMS_DIR}/bin/fake_declare_product ${product_dir} ${package} ${pkgver} ${fullqual}

setup -B ${package} ${pkgver} -q ${fullqual} -z ${fakedb}:${product_dir}:${PRODUCTS} || die "ERROR: fake setup failed"

if [[ "${basequal}" == e[67] ]]
then
  cpflg="-O2 -std=c11"
  cxxflg="-O2 -std=c++1y"
else
  die "Qualifier $basequal not recognized."
fi

set -x

gcc --version

cd ${blddir}

make OUT=${blddir} CPPFLAGS="${cpflg}"  VERBOSE=t src_lib src_utility  || die "ERROR: build of src_lib src_utility failed"
make OUT=${blddir} CPPFLAGS="${cxxflg}" VERBOSE=t src_example_user   || die "ERROR: build of src_example_user failed"
make OUT=${blddir} VERBOSE=t -j${ncore} modules                 || die "ERROR: build of modules failed"               

set +x

if [ ! -d ${pkgdir}/lib ]
then
   echo "ERROR: failed to create ${pkgdir}/lib"
   echo "       Something is very wrong"
   exit 1
fi

# real ups declare
${SSIBUILDSHIMS_DIR}/bin/declare_product ${product_dir} ${package} ${pkgver} ${fullqual}

# -------------------------------------------------------------------
# common bottom stuff
# -------------------------------------------------------------------

# this should not complain
echo "Finished building ${package} ${pkgver}"
setup ${package} ${pkgver} -q ${fullqual} -z ${product_dir}:${PRODUCTS}
echo "TRACE is installed at ${TRACE_FQ_DIR}"

# this must be last
if [ "${maketar}" = "tar" ] && [ -d ${pkgdir}/lib ]
then
   make_distribution_tarball
fi

exit 0
